# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the artifact service."""

import enum

import pytest
from google.genai import types

from adk_extra_services.artifacts import S3ArtifactService, LocalFolderArtifactService

Enum = enum.Enum


class ArtifactServiceType(Enum):
    S3 = "S3"
    LOCAL = "LOCAL"


def mock_s3_artifact_service():
    """Mocks an S3 client for testing S3ArtifactService."""

    class MockBody:

        def __init__(self, content):
            self._content = content

        def read(self):
            return self._content

    class MockS3Client:

        def __init__(self):
            self.store = {}

        class exceptions:

            class NoSuchKey(Exception):
                pass

        def put_object(self, Bucket, Key, Body, ContentType):
            self.store[Key] = (Body, ContentType)

        def get_object(self, Bucket, Key):
            if Key not in self.store:
                raise MockS3Client.exceptions.NoSuchKey()
            body, mime = self.store[Key]
            return {"Body": MockBody(body), "ContentType": mime}

        def delete_object(self, Bucket, Key):
            self.store.pop(Key, None)

        def get_paginator(self, operation_name):
            class Paginator:

                def paginate(self, Bucket, Prefix):
                    keys = [k for k in client.store.keys() if k.startswith(Prefix)]
                    yield {"Contents": [{"Key": k} for k in keys]}

            client = self
            return Paginator()

    svc = S3ArtifactService(bucket_name="test_bucket")
    svc.s3_client = MockS3Client()
    return svc


def get_artifact_service(service_type: ArtifactServiceType, tmp_path):
    """Returns an artifact service instance based on type."""
    if service_type == ArtifactServiceType.S3:
        return mock_s3_artifact_service()
    if service_type == ArtifactServiceType.LOCAL:
        return LocalFolderArtifactService(base_path=tmp_path)
    raise ValueError(f"Unsupported service type: {service_type}")


@pytest.mark.asyncio
@pytest.mark.parametrize("service_type", [ArtifactServiceType.S3, ArtifactServiceType.LOCAL])
async def test_load_empty(service_type, tmp_path):
    """Tests loading an artifact when none exists."""
    artifact_service = get_artifact_service(service_type, tmp_path)
    assert not await artifact_service.load_artifact(
        app_name="test_app",
        user_id="test_user",
        session_id="session_id",
        filename="filename",
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("service_type", [ArtifactServiceType.S3, ArtifactServiceType.LOCAL])
async def test_save_load_delete(service_type, tmp_path):
    """Tests saving, loading, and deleting an artifact."""
    artifact_service = get_artifact_service(service_type, tmp_path)
    artifact = types.Part.from_bytes(data=b"test_data", mime_type="text/plain")
    app_name = "app0"
    user_id = "user0"
    session_id = "123"
    filename = "file456"

    await artifact_service.save_artifact(
        app_name=app_name,
        user_id=user_id,
        session_id=session_id,
        filename=filename,
        artifact=artifact,
    )
    assert (
        await artifact_service.load_artifact(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=filename,
        )
        == artifact
    )

    await artifact_service.delete_artifact(
        app_name=app_name,
        user_id=user_id,
        session_id=session_id,
        filename=filename,
    )
    assert not await artifact_service.load_artifact(
        app_name=app_name,
        user_id=user_id,
        session_id=session_id,
        filename=filename,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("service_type", [ArtifactServiceType.S3, ArtifactServiceType.LOCAL])
async def test_list_keys(service_type, tmp_path):
    """Tests listing keys in the artifact service."""
    artifact_service = get_artifact_service(service_type, tmp_path)
    artifact = types.Part.from_bytes(data=b"test_data", mime_type="text/plain")
    app_name = "app0"
    user_id = "user0"
    session_id = "123"
    filename = "filename"
    filenames = [filename + str(i) for i in range(5)]

    for f in filenames:
        await artifact_service.save_artifact(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=f,
            artifact=artifact,
        )

    assert (
        await artifact_service.list_artifact_keys(
            app_name=app_name, user_id=user_id, session_id=session_id
        )
        == filenames
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("service_type", [ArtifactServiceType.S3, ArtifactServiceType.LOCAL])
async def test_list_versions(service_type, tmp_path):
    """Tests listing versions of an artifact."""
    artifact_service = get_artifact_service(service_type, tmp_path)

    app_name = "app0"
    user_id = "user0"
    session_id = "123"
    filename = "filename"
    versions = [
        types.Part.from_bytes(
            data=i.to_bytes(2, byteorder="big"), mime_type="text/plain"
        )
        for i in range(3)
    ]

    for i in range(3):
        await artifact_service.save_artifact(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=filename,
            artifact=versions[i],
        )

    response_versions = await artifact_service.list_versions(
        app_name=app_name,
        user_id=user_id,
        session_id=session_id,
        filename=filename,
    )

    assert response_versions == list(range(3))



"""Session service implementations for Google ADK."""

from .mongo_session_service import MongoSessionService
from .redis_session_service import RedisSessionService

__all__ = ["RedisSessionService", "MongoSessionService"]



"""Local folder artifact service implementation for Google ADK."""

from __future__ import annotations

import logging
import mimetypes
import os
from pathlib import Path
from typing import Optional

from google.adk.artifacts import BaseArtifactService
from google.genai import types
from typing_extensions import override

logger = logging.getLogger("adk_extra_services.artifacts.local_folder")


class LocalFolderArtifactService(BaseArtifactService):
    """Stores artifacts in the local filesystem."""

    def __init__(self, base_path: str | Path):
        self.base_path = Path(base_path).expanduser().resolve()
        self.base_path.mkdir(parents=True, exist_ok=True)

    def _file_has_user_namespace(self, filename: str) -> bool:
        return filename.startswith("user:")

    def _strip_user_prefix(self, filename: str) -> str:
        """Remove the leading ``user:`` namespace marker, if present."""
        return filename[len("user:") :] if self._file_has_user_namespace(filename) else filename

    def _get_file_path(
        self,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
        version: int,
    ) -> Path:
        """Translate a logical filename into an on-disk path."""
        if self._file_has_user_namespace(filename):
            clean_name = self._strip_user_prefix(filename)
            parts = [app_name, user_id, "user", clean_name, str(version)]
        else:
            parts = [app_name, user_id, session_id, filename, str(version)]
        return self.base_path.joinpath(*parts)

    @override
    async def save_artifact(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
        artifact: types.Part,
    ) -> int:
        versions = await self.list_versions(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=filename,
        )
        version = 0 if not versions else max(versions) + 1

        path = self._get_file_path(app_name, user_id, session_id, filename, version)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(artifact.inline_data.data)
        return version

    @override
    async def load_artifact(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
        version: Optional[int] = None,
    ) -> Optional[types.Part]:
        if version is None:
            versions = await self.list_versions(
                app_name=app_name,
                user_id=user_id,
                session_id=session_id,
                filename=filename,
            )
            if not versions:
                return None
            version = max(versions)

        path = self._get_file_path(app_name, user_id, session_id, filename, version)
        if not path.exists():
            return None

        data = path.read_bytes()
        mime, _ = mimetypes.guess_type(filename)
        if mime is None:
            mime = "text/plain"
        return types.Part.from_bytes(data=data, mime_type=mime)

    @override
    async def list_artifact_keys(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> list[str]:
        """Return all artifact filenames visible to the caller.

        This walks the on-disk directory tree and extracts the *logical*
        filenames by stripping the version component from the stored path and,
        for user-scoped artefacts, prepending the ``user:`` namespace.
        """
        keys: set[str] = set()

        def _collect(prefix: Path, add_user_prefix: bool = False) -> None:
            if not prefix.exists():
                return
            # Every stored artefact is saved under:
            #   <prefix>/<filename_path_components...>/<version>
            # where <version> is an integer file (not directory).  We therefore
            # look for *files* directly under any depth whose names are ints.
            for version_file in prefix.rglob("*"):
                if not version_file.is_file():
                    continue
                try:
                    int(version_file.name)
                except ValueError:
                    # Not a version file → skip
                    continue
                rel_parts = version_file.relative_to(prefix).parts[:-1]
                if not rel_parts:
                    # Malformed – skip
                    continue
                if rel_parts[0].startswith("user:"):
                    rel_parts = (rel_parts[0][len("user:"):],) + rel_parts[1:]
                logical_name = "/".join(rel_parts)
                if add_user_prefix:
                    logical_name = f"user:{logical_name}"
                keys.add(logical_name)

        # Session-scoped artefacts
        _collect(self.base_path / app_name / user_id / session_id)
        # User-scoped artefacts – prepend namespace
        _collect(self.base_path / app_name / user_id / "user", add_user_prefix=True)

        return sorted(keys)

    @override
    async def delete_artifact(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
    ) -> None:
        versions = await self.list_versions(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=filename,
        )
        for ver in versions:
            path = self._get_file_path(app_name, user_id, session_id, filename, ver)
            if path.exists():
                try:
                    path.unlink()
                except OSError as exc:
                    logger.warning("Failed to delete %s: %s", path, exc)

    @override
    async def list_versions(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
    ) -> list[int]:
        if self._file_has_user_namespace(filename):
            dir_path = self.base_path / app_name / user_id / "user" / self._strip_user_prefix(filename)
        else:
            dir_path = self.base_path / app_name / user_id / session_id / filename
        if not dir_path.exists():
            return []
        versions: list[int] = []
        for item in dir_path.iterdir():
            if item.is_file():
                try:
                    versions.append(int(item.name))
                except ValueError:
                    continue
        return sorted(versions)



"""S3 artifact service implementation for Google ADK."""

import logging
from typing import Optional

import boto3
from google.adk.artifacts import BaseArtifactService
from google.genai import types
from typing_extensions import override

logger = logging.getLogger("adk_extra_services.artifacts.s3")


class S3ArtifactService(BaseArtifactService):
    """An artifact service implementation using AWS S3 or S3-compatible storage."""

    def __init__(
        self,
        bucket_name: str,
        endpoint_url: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        region_name: Optional[str] = None,
        **kwargs,
    ):
        """Initializes the S3ArtifactService.

        Args:
            bucket_name: The name of the S3 bucket.
            endpoint_url: Optional endpoint URL for S3-compatible storage (e.g., MinIO).
            aws_access_key_id: Optional AWS access key. Uses environment/config if not provided.
            aws_secret_access_key: Optional AWS secret key. Uses environment/config if not provided.
            region_name: Optional AWS region. Uses environment/config if not provided.
            **kwargs: Additional keyword arguments to pass to boto3.client('s3').
        """
        self.bucket_name = bucket_name

        # Prepare S3 client configuration
        client_kwargs = kwargs.copy()
        if endpoint_url:
            client_kwargs["endpoint_url"] = endpoint_url
        if aws_access_key_id:
            client_kwargs["aws_access_key_id"] = aws_access_key_id
        if aws_secret_access_key:
            client_kwargs["aws_secret_access_key"] = aws_secret_access_key
        if region_name:
            client_kwargs["region_name"] = region_name

        self.s3_client = boto3.client("s3", **client_kwargs)

    def _file_has_user_namespace(self, filename: str) -> bool:
        return filename.startswith("user:")

    def _get_object_key(
        self,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
        version: int,
    ) -> str:
        if self._file_has_user_namespace(filename):
            return f"{app_name}/{user_id}/user/{filename}/{version}"
        return f"{app_name}/{user_id}/{session_id}/{filename}/{version}"

    @override
    async def save_artifact(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
        artifact: types.Part,
    ) -> int:
        versions = await self.list_versions(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=filename,
        )
        version = 0 if not versions else max(versions) + 1

        key = self._get_object_key(app_name, user_id, session_id, filename, version)
        self.s3_client.put_object(
            Bucket=self.bucket_name,
            Key=key,
            Body=artifact.inline_data.data,
            ContentType=artifact.inline_data.mime_type,
        )
        return version

    @override
    async def load_artifact(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
        version: Optional[int] = None,
    ) -> Optional[types.Part]:
        if version is None:
            versions = await self.list_versions(
                app_name=app_name,
                user_id=user_id,
                session_id=session_id,
                filename=filename,
            )
            if not versions:
                return None
            version = max(versions)

        key = self._get_object_key(app_name, user_id, session_id, filename, version)
        try:
            resp = self.s3_client.get_object(Bucket=self.bucket_name, Key=key)
            data = resp["Body"].read()
            mime = resp.get("ContentType")
        except self.s3_client.exceptions.NoSuchKey:
            return None

        return types.Part.from_bytes(data=data, mime_type=mime)

    @override
    async def list_artifact_keys(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> list[str]:
        filenames = set()

        session_prefix = f"{app_name}/{user_id}/{session_id}/"
        paginator = self.s3_client.get_paginator("list_objects_v2")
        for page in paginator.paginate(Bucket=self.bucket_name, Prefix=session_prefix):
            for obj in page.get("Contents", []):
                parts = obj["Key"].split("/")
                if len(parts) >= 5:
                    filenames.add(parts[3])

        user_prefix = f"{app_name}/{user_id}/user/"
        for page in paginator.paginate(Bucket=self.bucket_name, Prefix=user_prefix):
            for obj in page.get("Contents", []):
                parts = obj["Key"].split("/")
                if len(parts) >= 5:
                    filenames.add(parts[3])

        return sorted(filenames)

    @override
    async def delete_artifact(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
    ) -> None:
        versions = await self.list_versions(
            app_name=app_name,
            user_id=user_id,
            session_id=session_id,
            filename=filename,
        )
        for ver in versions:
            key = self._get_object_key(app_name, user_id, session_id, filename, ver)
            self.s3_client.delete_object(Bucket=self.bucket_name, Key=key)

    @override
    async def list_versions(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        filename: str,
    ) -> list[int]:
        if self._file_has_user_namespace(filename):
            prefix = f"{app_name}/{user_id}/user/{filename}/"
        else:
            prefix = f"{app_name}/{user_id}/{session_id}/{filename}/"
        versions = []
        paginator = self.s3_client.get_paginator("list_objects_v2")
        try:
            for page in paginator.paginate(Bucket=self.bucket_name, Prefix=prefix):
                for obj in page.get("Contents", []):
                    parts = obj["Key"].rstrip("/").split("/")
                    ver_str = parts[-1]
                    try:
                        versions.append(int(ver_str))
                    except ValueError:
                        continue
        except self.s3_client.exceptions.NoSuchKey:
            # Happens when prefix does not yet exist in the bucket. Treat as no versions.
            return []
        except self.s3_client.exceptions.NoSuchBucket:
            logger.error("Bucket %s does not exist", self.bucket_name)
            raise

        return versions



"""MongoDB session service implementation for Google ADK."""

import copy
import time
import uuid
from typing import Any, Optional

import motor.motor_asyncio
from google.adk.events.event import Event
from google.adk.sessions.base_session_service import (
    BaseSessionService,
    GetSessionConfig,
    ListSessionsResponse,
)
from google.adk.sessions.session import Session
from google.adk.sessions.state import State


class MongoSessionService(BaseSessionService):

    def __init__(self, mongo_url: str, db_name: Optional[str] = None, **kwargs: Any):
        client = motor.motor_asyncio.AsyncIOMotorClient(mongo_url, **kwargs)
        self.db = client[db_name] if db_name else client.get_default_database()
        self.sessions = self.db.sessions
        self.events = self.db.events
        self.app_states = self.db.app_states
        self.user_states = self.db.user_states

    async def create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        sid = session_id or uuid.uuid4().hex
        now = time.time()
        session_state = state or {}

        await self.sessions.insert_one(
            {
                "app_name": app_name,
                "user_id": user_id,
                "id": sid,
                "state": session_state,
                "last_update_time": now,
            }
        )

        session = Session(
            id=sid,
            app_name=app_name,
            user_id=user_id,
            state=session_state,
            events=[],
            last_update_time=now,
        )
        return await self._merge_state(app_name, user_id, copy.deepcopy(session))

    async def get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        filt = {"app_name": app_name, "user_id": user_id, "id": session_id}
        doc = await self.sessions.find_one(filt)
        if not doc:
            return None

        # Get all events for this session
        raw_events = (
            await self.events.find(filt).sort("timestamp", 1).to_list(length=None)
        )
        events = [Event.model_validate_json(e["raw"]) for e in raw_events]

        # Apply config filters if provided
        if config:
            if config.after_timestamp is not None:
                events = [e for e in events if e.timestamp >= config.after_timestamp]
            if config.num_recent_events is not None and events:
                events = events[-config.num_recent_events :]

        session = Session(
            id=session_id,
            app_name=app_name,
            user_id=user_id,
            state=doc.get("state", {}),
            events=events,
            last_update_time=doc.get("last_update_time", 0.0),
        )

        return await self._merge_state(app_name, user_id, copy.deepcopy(session))

    async def list_sessions(
        self, *, app_name: str, user_id: str
    ) -> ListSessionsResponse:
        cursor = self.sessions.find({"app_name": app_name, "user_id": user_id})
        sessions: list[Session] = []
        async for doc in cursor:
            sessions.append(
                Session(
                    id=doc["id"],
                    app_name=app_name,
                    user_id=user_id,
                    state={},
                    events=[],
                    last_update_time=doc.get("last_update_time", 0.0),
                )
            )
        return ListSessionsResponse(sessions=sessions)

    async def delete_session(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> None:
        filt = {"app_name": app_name, "user_id": user_id, "id": session_id}
        await self.sessions.delete_one(filt)
        await self.events.delete_many(filt)

    async def _merge_state(
        self, app_name: str, user_id: str, session: Session
    ) -> Session:
        """Merge app and user state into the session."""
        # Merge app state
        app_states = await self.app_states.find({"app_name": app_name}).to_list(
            length=None
        )
        for state in app_states:
            session.state[State.APP_PREFIX + state["key"]] = state["value"]

        # Merge user state
        user_states = await self.user_states.find(
            {"app_name": app_name, "user_id": user_id}
        ).to_list(length=None)
        for state in user_states:
            session.state[State.USER_PREFIX + state["key"]] = state["value"]

        return session

    async def append_event(self, session: Session, event: Event) -> Event:
        if event.partial:
            return event

        filt = {
            "app_name": session.app_name,
            "user_id": session.user_id,
            "id": session.id,
        }
        doc = await self.sessions.find_one(filt)
        if not doc:
            raise ValueError("session not found")

        if doc.get("last_update_time", 0.0) > session.last_update_time:
            raise ValueError("stale session")

        new_event = await super().append_event(session=session, event=event)

        # Store the event
        await self.events.insert_one(
            {
                **filt,
                "raw": new_event.model_dump_json(),
                "timestamp": new_event.timestamp,
            }
        )

        # Update session state and timestamp
        await self.sessions.update_one(
            filt,
            {
                "$set": {
                    "state": session.state,
                    "last_update_time": session.last_update_time,
                }
            },
        )

        # Process state deltas if present
        if event.actions and event.actions.state_delta:
            for key, value in event.actions.state_delta.items():
                if key.startswith(State.APP_PREFIX):
                    # Update app state
                    app_key = key.removeprefix(State.APP_PREFIX)
                    await self.app_states.update_one(
                        {"app_name": session.app_name, "key": app_key},
                        {"$set": {"value": value}},
                        upsert=True,
                    )

                elif key.startswith(State.USER_PREFIX):
                    # Update user state
                    user_key = key.removeprefix(State.USER_PREFIX)
                    await self.user_states.update_one(
                        {
                            "app_name": session.app_name,
                            "user_id": session.user_id,
                            "key": user_key,
                        },
                        {"$set": {"value": value}},
                        upsert=True,
                    )

        return new_event



"""Artifact service implementations for Google ADK."""

from .s3_artifact_service import S3ArtifactService
from .local_folder_artifact_service import LocalFolderArtifactService

__all__ = [
    "S3ArtifactService",
    "LocalFolderArtifactService",
]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum

import pytest
from google.adk.events import Event, EventActions
from google.adk.sessions.base_session_service import GetSessionConfig
from google.genai import types
from testcontainers.mongodb import MongoDbContainer
from testcontainers.redis import RedisContainer

from adk_extra_services.sessions import MongoSessionService, RedisSessionService

MONGO_URL = None
REDIS_URL = None


class SessionServiceType(enum.Enum):
    MONGO = "MONGO"
    REDIS = "REDIS"


@pytest.fixture(scope="session")
def mongo_url():
    with MongoDbContainer() as container:
        yield container.get_connection_url()


@pytest.fixture(scope="session")
def redis_url():
    with RedisContainer() as container:
        host = container.get_container_host_ip()
        port = container.get_exposed_port(6379)
        yield f"redis://{host}:{port}"


@pytest.fixture(scope="session", autouse=True)
def set_urls(mongo_url, redis_url):
    global MONGO_URL, REDIS_URL
    MONGO_URL = mongo_url
    REDIS_URL = redis_url


def get_session_service(
    service_type: SessionServiceType,
):
    """Creates a session service for testing."""
    if service_type == SessionServiceType.MONGO:
        return MongoSessionService(MONGO_URL, db_name="test")
    if service_type == SessionServiceType.REDIS:
        return RedisSessionService(REDIS_URL)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_get_empty_session(service_type):
    session_service = get_session_service(service_type)
    assert not await session_service.get_session(
        app_name="my_app", user_id="test_user", session_id="123"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_create_get_session(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id = "test_user"
    state = {"key": "value"}

    session = await session_service.create_session(
        app_name=app_name, user_id=user_id, state=state
    )
    assert session.app_name == app_name
    assert session.user_id == user_id
    assert session.id
    assert session.state == state
    assert (
        await session_service.get_session(
            app_name=app_name, user_id=user_id, session_id=session.id
        )
        == session
    )

    session_id = session.id
    await session_service.delete_session(
        app_name=app_name, user_id=user_id, session_id=session_id
    )

    assert (
        await session_service.get_session(
            app_name=app_name, user_id=user_id, session_id=session.id
        )
        != session
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_create_and_list_sessions(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id = "test_user"

    session_ids = ["session" + str(i) for i in range(5)]
    for session_id in session_ids:
        await session_service.create_session(
            app_name=app_name, user_id=user_id, session_id=session_id
        )

    list_sessions_response = await session_service.list_sessions(
        app_name=app_name, user_id=user_id
    )
    sessions = list_sessions_response.sessions
    for i in range(len(sessions)):
        assert sessions[i].id == session_ids[i]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_session_state(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id_1 = "user1"
    user_id_2 = "user2"
    session_id_11 = "session11"
    session_id_12 = "session12"
    session_id_2 = "session2"
    state_11 = {"key11": "value11"}
    state_12 = {"key12": "value12"}

    session_11 = await session_service.create_session(
        app_name=app_name,
        user_id=user_id_1,
        state=state_11,
        session_id=session_id_11,
    )
    await session_service.create_session(
        app_name=app_name,
        user_id=user_id_1,
        state=state_12,
        session_id=session_id_12,
    )
    await session_service.create_session(
        app_name=app_name, user_id=user_id_2, session_id=session_id_2
    )

    assert session_11.state.get("key11") == "value11"

    event = Event(
        invocation_id="invocation",
        author="user",
        content=types.Content(role="user", parts=[types.Part(text="text")]),
        actions=EventActions(
            state_delta={
                "app:key": "value",
                "user:key1": "value1",
                "temp:key": "temp",
                "key11": "value11_new",
            }
        ),
    )
    await session_service.append_event(session=session_11, event=event)

    # User and app state is stored, temp state is filtered.
    assert session_11.state.get("app:key") == "value"
    assert session_11.state.get("key11") == "value11_new"
    assert session_11.state.get("user:key1") == "value1"
    assert not session_11.state.get("temp:key")

    session_12 = await session_service.get_session(
        app_name=app_name, user_id=user_id_1, session_id=session_id_12
    )
    # After getting a new instance, the session_12 got the user and app state,
    # even append_event is not applied to it, temp state has no effect
    assert session_12.state.get("key12") == "value12"
    assert not session_12.state.get("temp:key")

    # The user1's state is not visible to user2, app state is visible
    session_2 = await session_service.get_session(
        app_name=app_name, user_id=user_id_2, session_id=session_id_2
    )
    assert session_2.state.get("app:key") == "value"
    assert not session_2.state.get("user:key1")

    assert not session_2.state.get("user:key1")

    # The change to session_11 is persisted
    session_11 = await session_service.get_session(
        app_name=app_name, user_id=user_id_1, session_id=session_id_11
    )
    assert session_11.state.get("key11") == "value11_new"
    assert session_11.state.get("user:key1") == "value1"
    assert not session_11.state.get("temp:key")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_create_new_session_will_merge_states(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id = "user"
    session_id_1 = "session1"
    session_id_2 = "session2"
    state_1 = {"key1": "value1"}

    session_1 = await session_service.create_session(
        app_name=app_name, user_id=user_id, state=state_1, session_id=session_id_1
    )

    event = Event(
        invocation_id="invocation",
        author="user",
        content=types.Content(role="user", parts=[types.Part(text="text")]),
        actions=EventActions(
            state_delta={
                "app:key": "value",
                "user:key1": "value1",
                "temp:key": "temp",
            }
        ),
    )
    await session_service.append_event(session=session_1, event=event)

    # User and app state is stored, temp state is filtered.
    assert session_1.state.get("app:key") == "value"
    assert session_1.state.get("key1") == "value1"
    assert session_1.state.get("user:key1") == "value1"
    assert not session_1.state.get("temp:key")

    session_2 = await session_service.create_session(
        app_name=app_name, user_id=user_id, state={}, session_id=session_id_2
    )
    # Session 2 has the persisted states
    assert session_2.state.get("app:key") == "value"
    assert session_2.state.get("user:key1") == "value1"
    assert not session_2.state.get("key1")
    assert not session_2.state.get("temp:key")


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_append_event_bytes(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id = "user"

    session = await session_service.create_session(app_name=app_name, user_id=user_id)

    test_content = types.Content(
        role="user",
        parts=[
            types.Part.from_bytes(data=b"test_image_data", mime_type="image/png"),
        ],
    )
    test_grounding_metadata = types.GroundingMetadata(
        search_entry_point=types.SearchEntryPoint(sdk_blob=b"test_sdk_blob")
    )
    event = Event(
        invocation_id="invocation",
        author="user",
        content=test_content,
        grounding_metadata=test_grounding_metadata,
    )
    await session_service.append_event(session=session, event=event)

    assert session.events[0].content == test_content

    session = await session_service.get_session(
        app_name=app_name, user_id=user_id, session_id=session.id
    )
    events = session.events
    assert len(events) == 1
    assert events[0].content == test_content
    assert events[0].grounding_metadata == test_grounding_metadata


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_append_event_complete(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id = "user"

    session = await session_service.create_session(app_name=app_name, user_id=user_id)
    event = Event(
        invocation_id="invocation",
        author="user",
        content=types.Content(role="user", parts=[types.Part(text="test_text")]),
        turn_complete=True,
        partial=False,
        actions=EventActions(
            artifact_delta={
                "file": 0,
            },
            transfer_to_agent="agent",
            escalate=True,
        ),
        long_running_tool_ids={"tool1"},
        error_code="error_code",
        error_message="error_message",
        interrupted=True,
    )
    await session_service.append_event(session=session, event=event)

    assert (
        await session_service.get_session(
            app_name=app_name, user_id=user_id, session_id=session.id
        )
        == session
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "service_type", [SessionServiceType.MONGO, SessionServiceType.REDIS]
)
async def test_get_session_with_config(service_type):
    session_service = get_session_service(service_type)
    app_name = "my_app"
    user_id = "user"

    num_test_events = 5
    session = await session_service.create_session(app_name=app_name, user_id=user_id)
    for i in range(1, num_test_events + 1):
        event = Event(author="user", timestamp=i)
        await session_service.append_event(session, event)

    # No config, expect all events to be returned.
    session = await session_service.get_session(
        app_name=app_name, user_id=user_id, session_id=session.id
    )
    events = session.events
    assert len(events) == num_test_events

    # Only expect the most recent 3 events.
    num_recent_events = 3
    config = GetSessionConfig(num_recent_events=num_recent_events)
    session = await session_service.get_session(
        app_name=app_name, user_id=user_id, session_id=session.id, config=config
    )
    events = session.events
    assert len(events) == num_recent_events
    assert events[0].timestamp == num_test_events - num_recent_events + 1

    # Only expect events after timestamp 4.0 (inclusive), i.e., 2 events.
    after_timestamp = 4.0
    config = GetSessionConfig(after_timestamp=after_timestamp)
    session = await session_service.get_session(
        app_name=app_name, user_id=user_id, session_id=session.id, config=config
    )
    events = session.events
    assert len(events) == num_test_events - after_timestamp + 1
    assert events[0].timestamp == after_timestamp

    # Expect no events if none are > after_timestamp.
    way_after_timestamp = num_test_events * 10
    config = GetSessionConfig(after_timestamp=way_after_timestamp)
    session = await session_service.get_session(
        app_name=app_name, user_id=user_id, session_id=session.id, config=config
    )
    assert not session.events

    # Both filters applied, i.e., of 3 most recent events, only 2 are after
    # timestamp 4.0, so expect 2 events.
    config = GetSessionConfig(
        after_timestamp=after_timestamp, num_recent_events=num_recent_events
    )
    session = await session_service.get_session(
        app_name=app_name, user_id=user_id, session_id=session.id, config=config
    )
    events = session.events
    assert len(events) == num_test_events - after_timestamp + 1



"""Redis session service implementation for Google ADK."""

import json
import logging
import time
import uuid
from typing import Any, Dict, Optional

import redis.asyncio as aioredis
from google.adk.events.event import Event
from google.adk.sessions.base_session_service import (
    BaseSessionService,
    GetSessionConfig,
    ListSessionsResponse,
)
from google.adk.sessions.session import Session
from google.adk.sessions.state import State


def _meta_key(app: str, user: str, session: str) -> str:
    return f"adk:sessions:{app}:{user}:{session}:meta"


def _state_key(app: str, user: str, session: str) -> str:
    return f"adk:sessions:{app}:{user}:{session}:state"


def _events_key(app: str, user: str, session: str) -> str:
    return f"adk:sessions:{app}:{user}:{session}:events"


def _user_set_key(app: str, user: str) -> str:
    return f"adk:sessions:{app}:{user}:sessions"


def _app_state_key(app: str) -> str:
    return f"adk:sessions:{app}:app_state"


def _user_state_key(app: str, user: str) -> str:
    return f"adk:sessions:{app}:{user}:user_state"


class RedisSessionService(BaseSessionService):

    def __init__(self, redis_url: str, **kwargs: Any):
        self.logger = logging.getLogger(__name__)
        self.client = aioredis.from_url(redis_url, **kwargs)

    async def create_session(
        self,
        *,
        app_name: str,
        user_id: str,
        state: Optional[dict[str, Any]] = None,
        session_id: Optional[str] = None,
    ) -> Session:
        sid = session_id or uuid.uuid4().hex
        now = time.time()
        await self.client.hset(
            _meta_key(app_name, user_id, sid),
            mapping={"id": sid, "last_update_time": now},
        )
        await self.client.set(
            _state_key(app_name, user_id, sid), json.dumps(state or {})
        )
        await self.client.delete(_events_key(app_name, user_id, sid))
        await self.client.sadd(_user_set_key(app_name, user_id), sid)

        # Create a session and merge state
        session = Session(
            id=sid,
            app_name=app_name,
            user_id=user_id,
            state=state or {},
            events=[],
            last_update_time=now,
        )
        return await self._merge_state(app_name, user_id, session)

    async def get_session(
        self,
        *,
        app_name: str,
        user_id: str,
        session_id: str,
        config: Optional[GetSessionConfig] = None,
    ) -> Optional[Session]:
        key = _meta_key(app_name, user_id, session_id)
        if not await self.client.exists(key):
            return None
        meta = await self.client.hgetall(key)
        last = float(meta.get(b"last_update_time", b"0"))
        state = json.loads(
            (await self.client.get(_state_key(app_name, user_id, session_id))) or b"{}"
        )
        raw = await self.client.lrange(
            _events_key(app_name, user_id, session_id), 0, -1
        )
        events = [Event.model_validate_json(e.decode()) for e in raw]

        # Apply config filters correctly
        if config:
            if config.after_timestamp is not None:
                # Use >= instead of > to match the expected behavior in tests
                events = [e for e in events if e.timestamp >= config.after_timestamp]
            if config.num_recent_events is not None:
                events = events[-config.num_recent_events :]

        session = Session(
            id=session_id,
            app_name=app_name,
            user_id=user_id,
            state=state,
            events=events,
            last_update_time=last,
        )
        return await self._merge_state(app_name, user_id, session)

    async def list_sessions(
        self, *, app_name: str, user_id: str
    ) -> ListSessionsResponse:
        ids = await self.client.smembers(_user_set_key(app_name, user_id))
        sessions: list[Session] = []
        # Sort the session IDs to ensure consistent ordering for tests
        sorted_ids = sorted([sid.decode() for sid in ids])
        for sid_str in sorted_ids:
            last_b = (
                await self.client.hget(
                    _meta_key(app_name, user_id, sid_str), "last_update_time"
                )
                or b"0"
            )
            last = float(last_b)
            sessions.append(
                Session(
                    id=sid_str,
                    app_name=app_name,
                    user_id=user_id,
                    state={},
                    events=[],
                    last_update_time=last,
                )
            )
        return ListSessionsResponse(sessions=sessions)

    async def delete_session(
        self, *, app_name: str, user_id: str, session_id: str
    ) -> None:
        keys = [
            _meta_key(app_name, user_id, session_id),
            _state_key(app_name, user_id, session_id),
            _events_key(app_name, user_id, session_id),
        ]
        await self.client.delete(*keys)
        await self.client.srem(_user_set_key(app_name, user_id), session_id)

    async def append_event(self, session: Session, event: Event) -> Event:
        if event.partial:
            return event
        mkey = _meta_key(session.app_name, session.user_id, session.id)
        stored = await self.client.hget(mkey, "last_update_time") or b"0"
        if float(stored) > session.last_update_time:
            raise ValueError("stale session")

        # Process the event using the parent class implementation
        new_event = await super().append_event(session=session, event=event)

        # Update user and app state if there's a state delta
        if event.actions and event.actions.state_delta:
            for key, value in event.actions.state_delta.items():
                if key.startswith(State.APP_PREFIX):
                    app_key = key.removeprefix(State.APP_PREFIX)
                    await self.client.hset(
                        _app_state_key(session.app_name), app_key, json.dumps(value)
                    )
                elif key.startswith(State.USER_PREFIX):
                    user_key = key.removeprefix(State.USER_PREFIX)
                    await self.client.hset(
                        _user_state_key(session.app_name, session.user_id),
                        user_key,
                        json.dumps(value),
                    )

        # Save the event and update session state
        await self.client.rpush(
            _events_key(session.app_name, session.user_id, session.id),
            new_event.model_dump_json(),
        )
        await self.client.set(
            _state_key(session.app_name, session.user_id, session.id),
            json.dumps(session.state),
        )
        await self.client.hset(mkey, "last_update_time", session.last_update_time)

        return new_event

    async def _merge_state(
        self, app_name: str, user_id: str, session: Session
    ) -> Session:
        """Merge app and user state into the session state."""
        # Merge app state
        app_state = await self.client.hgetall(_app_state_key(app_name))
        for key, value_json in app_state.items():
            key_str = key.decode() if isinstance(key, bytes) else key
            value = json.loads(
                value_json.decode() if isinstance(value_json, bytes) else value_json
            )
            session.state[State.APP_PREFIX + key_str] = value

        # Merge user state
        user_state = await self.client.hgetall(_user_state_key(app_name, user_id))
        for key, value_json in user_state.items():
            key_str = key.decode() if isinstance(key, bytes) else key
            value = json.loads(
                value_json.decode() if isinstance(value_json, bytes) else value_json
            )
            session.state[State.USER_PREFIX + key_str] = value

        return session
